unit MainForm;

interface

uses Windows, Messages, NUtil, NForm, NCmd, NkVImage, ImgFilter, SpiFilter,
  InfoDialog;

var
  AppName, ProfDir, About: string;

type
  TFileList = class(TStrList)
    Cur: Integer;
  end;  

  TNkVMainForm = class(TForm)
    Cmd: TCmdList;
    Image: TNkVImage;
    InfoDlg: TInfoDlg;
    FileList: TFileList;
    FileListUpdate: Boolean;
    Archiv: TArchive;
    ArFileList: TFileList;
    DirMode: Boolean;
    Stretch: Boolean;
    ViewLeft, ViewTop: Integer;
    DispWidth, DispHeight: Integer;
    Dw, Dh: Integer;
    ShowGrid: Boolean;
    ScaleFilter: Boolean;
    Plugins: TStrList;
    ArrowCursor: HCURSOR;
    HandCursor: HCURSOR;
    GrabCursor: HCURSOR;
    OldPt: TPoint;
    Border: Boolean;
    ShowColorValue: Boolean;
    FormList: TList;
    constructor Create(const Caption: string; ALeft, ATop, AWidth,
      AHeight: Integer); override;
    destructor Destroy; override;
    procedure Zoom(Value: Double);
    procedure Scroll(X, Y: Integer);
    procedure LoadConfig;
    procedure LoadPluginList;
    procedure LoadKeyMap;
    function LookupFilter(Buf: PChar; BufSize: Integer): TObject;
    procedure LoadFirstFile(const List: TFileList);
    procedure LoadFile(const Path: string; Last: Boolean);
    procedure LoadFirstFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadLastFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadFileFromArchive(const Ar: TArchive; const FileName: string);
    procedure UpdateView(zoom: Double);
    procedure SetClipboardText(const S: string);
    procedure SaveImage;
    procedure NextImage(const List: TFileList);
    procedure PrevImage(const List: TFileList; Last: Boolean);
    procedure DoCmd(const Name: string);
    procedure UpdateCaption;
    procedure UpdateCursor;
    procedure UpdateInfo;
    procedure FillFileList;
    procedure Error(const Path: string; const Msg: string);
    procedure SetWallpaper(Tile: Boolean);
    procedure ShowPluginConfig(const FileName: string);
    procedure ShowAboutPlugins;
    procedure ActivateNextForm(Prev: Boolean);
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMLButtonDown(var Message: TWMLButtonDown);message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk);
      message WM_LBUTTONDBLCLK;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMMouseWheel(var Message: TMessage); message WM_MOUSEWHEEL;
    procedure WMClose(var Message: TWMClose); message WM_CLOSE;
    procedure WMQueryEndSession(var Message: TWMQueryEndSession);
        message WM_QUERYENDSESSION;
  end;

implementation

uses ShellAPI, NSys, NLib, NConf, Command, ifdib, ifpnm;

const
  ZoomTable: array [0..15] of Double = (
    0.125, 0.25, 0.5, 0.667, 1.0, 1.25, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0,
    12.0, 16.0
  );
  IDM_PLUGINSETTING_FIRST = 4000;
  IDM_PLUGINSETTING_LAST = IDM_PLUGINSETTING_FIRST + 999;

function RoundOff(Value: Double): Longint;
begin
  if Value >= 0 then
    Result := Trunc(Value + 0.5)
  else
    Result := Trunc(Value - 0.5);
end;

function TNkVMainForm.LookupFilter(Buf: PChar; BufSize: Integer): TObject;
const
  Filters: array[0..1] of TImgFilterClass = (TDibFilter, TPnmFilter);
var
  i: Integer;
begin
  for i := 0 to High(Filters) do
  begin
    Result := Filters[i].Create;
    if (Result as TImgFilter).Test(Buf, BufSize) then
      Exit
    else
      Result.Free;
  end;

  for i := 0 to Plugins.Count-1 do
  begin
    Result := TSpiFilter.CreateByFile(Plugins[i]);
    if (Result as TImgFilter).Test(Buf, BufSize) then
    begin
      if (Result is TSpiFilter) and (Result as TSpiFilter).IsArchive then
      begin
        Result.Free;
        Result := TArchive.Create(Plugins[i]);
      end;
      Exit
    end else
      Result.Free;
  end;

  Result := nil;
end;

constructor TNkVMainForm.Create(const Caption: string; ALeft, ATop, AWidth,
  AHeight: Integer);
begin
  inherited;
  Border := True;
  ArrowCursor := LoadCursor(0, IDC_ARROW);
  HandCursor := LoadCursor(HInstance, 'HANDOPEN');
  GrabCursor := LoadCursor(HInstance, 'HANDHOLD');
  DragAcceptFiles(Handle, True);
  SetBackground(GetStockObject(NULL_BRUSH));
  Cmd := TCmdList.Create(CmdList);
  LoadKeyMap;
  SetAccelerator(Cmd.AccelHandle);
  Plugins := TStrList.Create; 
  LoadPluginList;
  Image := TNkVImage.Create;
  InfoDlg := TInfoDlg.CreateByResource(Handle, 'INFODIALOG');
  FileList := TFileList.Create;
  ArFileList := TFileList.Create;
end;

destructor TNkVMainForm.Destroy;
begin
  Cmd.Free;
  FileList.Free;
  ArFileList.Free;
  Plugins.Free;
  InfoDlg.Free;
end;

procedure TNkVMainForm.LoadPluginList;
var
  fname, dummy, ext: string;
begin
  try
    with TConf.Create(JoinPath(ProfDir, 'pluginlist')) do
      try
        while ReadLine do
          Plugins.Add(Line);
      finally
        Free;
      end;
  except
    on E: EIOError do
      with TListDir.Create(DirName(Argv[0])) do
        try
          while Each(fname) do
          begin
            SplitExt(fname, dummy, ext); 
            if not IsDir and (Lower(ext) = '.spi') then
              Plugins.Add(fname);
          end;
        finally
          Free;
        end;
  end;
end;

procedure TNkVMainForm.LoadConfig;
var
  s: string;
  w, h: Integer;
begin
  w := 320;
  h := 240;
  try
    with TConf.Create(JoinPath(ProfDir, 'config')) do
      try
        while ReadLine do
        begin
          s := GetToken;
          if s = cmStretchMode then
            Stretch := GetToken = 'on'
          else if s = cmScaleFilter then
            ScaleFilter := GetToken = 'on'
          else if s = 'width' then
            w := SInt(GetToken)
          else if s = 'height' then
            h := SInt(GetToken);
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do ;
  end;
  SetSize(w, h);
end;

procedure TNkVMainForm.LoadKeyMap;
var
  i: Integer;
  s: array[0..255] of string;
begin
  for i := 0 to High(s) do
    s[i] := '';
  i := 0;
  try
    with TConf.Create(JoinPath(ProfDir, 'keymap')) do
      try
        while ReadLine and (i < High(s)) do
        begin
          s[i] := GetToken;
          s[i+1] := GetToken;
          Inc(i, 2);
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do ;
  end;
  Cmd.BindKeys(DefaultKeys, s);
end;

procedure TNkVMainForm.FillFileList;
var
  s, dir, fname: string;
  i: Integer;
begin
  if not FileListUpdate then
    Exit;

  FileListUpdate := False;
  s := FileList[FileList.Cur];
  dir := DirName(s);
  FileList.Clear;
  with TListDir.Create(dir) do
    try
      while Each(fname) do
        if not IsDir then
          FileList.Add(JoinPath(dir, fname));
    finally
      Free;
    end;
  FileList.Sort(StrColl);
  
  for i := 0 to FileList.Count-1 do 
    if FileList[i] = s then
    begin
      FileList.Cur := i;
      Exit;
    end;
end;

procedure TNkVMainForm.LoadFirstFile(const List: TFileList);
var
  quiet: Boolean;
begin
  quiet := List.Count > 1;
  while List.Count > 0 do
  begin
    try
      LoadFile(List[0], False)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        if not quiet then
          Error(List[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    FileList.Assign(List);
    FileList.Cur := 0;
    UpdateCaption;
  end else
    Error('', 'Loadable file not found.');
end;

procedure TNkVMainForm.UpdateCursor;
var
  cursor: HCURSOR;
begin
  if ShowColorValue then
    cursor := LoadCursor(0, IDC_CROSS)
  else if Border then
    cursor := HandCursor
  else
    cursor := ArrowCursor;
  SetClassLong(Handle, GCL_HCURSOR, cursor);
  SetCursor(cursor);
end;

procedure TNkVMainForm.UpdateView(zoom: Double);
begin
  ViewLeft := 0;
  ViewTop := 0;
  DispWidth := RoundOff(Image.Width * zoom);
  DispHeight := RoundOff(Image.Height * zoom);
  Invalidate;
  Update;
  if not Stretch and not FullScreen and not Maximized then
    SetClientSize(DispWidth, DispHeight);
  UpdateCursor;
end;

procedure TNkVMainForm.LoadFileFromArchive(const Ar: TArchive;
  const FileName: string);
const
  TestBufSize = 2048;
var
  buf: PChar;
  fsize: Cardinal;
  filter: TObject;
  zoom: Double;
  oldcur: HCURSOR;
begin
  if Image.Loaded then
    zoom := DispWidth / Image.Width
  else
    zoom := 1.0;

  oldcur := SetCursor(LoadCursor(0, IDC_WAIT));
  fsize := Ar[FileName].FileSize;
  buf := nil;
  try
    GetMem(buf, fsize+TestBufSize);
    FillChar(buf^, TestBufSize, 0);
    Ar.Extract(FileName, buf, fsize);
    filter := LookupFilter(buf, fsize);
    try
      if not Assigned(filter) or (filter is TArchive) then
        raise EImgFilterError.Create('Unknown format.');
      Image.ReadMem(buf, fsize, filter as TImgFilter);
    finally
      filter.Free;
    end;
  finally
    FreeMem(buf);
    SetCursor(oldcur);
  end;
   
  // loding succeeded
  UpdateView(zoom);
end;

procedure TNkVMainForm.LoadFirstFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List[0])
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := 0;
    //UpdateCaption;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadLastFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List[List.Count-1])
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[List.Count-1], E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := List.Count-1;
    //UpdateCaption;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadFile(const Path: string; Last: Boolean);
const
  TestBufSize = 2048;
  StdinBufSize = 32768;
var
  len, fsize: Cardinal;
  buf: PChar;
  filter: TObject;
  f: TFile;
  isstdin: Boolean;
  oldcur: HCURSOR;
  zoom: Double;
  ar: TArchive;
  arlist: TFileList;
  fname: string;
begin
  if Image.Loaded then
    zoom := DispWidth / Image.Width
  else
    zoom := 1.0;

  isstdin := Path = '';
  if isstdin then
    f := TFile.CreateStdIn
  else
    f := TFile.Create(Path, 'r');

  buf := nil;
  oldcur := SetCursor(LoadCursor(0, IDC_WAIT));
  with f do
    try
      GetMem(buf, TestBufSize);
      FillChar(buf^, TestBufSize, 0);
      len := Read(buf^, TestBufSize);
      if isstdin and (len = 0) then
        Exit;
      filter := LookupFilter(buf, len);
      try
        if not Assigned(filter) then
          raise EImgFilterError.Create('Unknown format.');

        if filter is TArchive then
        begin
          ar := filter as TArchive;
          ar.Open(Path);
          arlist := TFileList.Create;
          while ar.EachName(fname) do
            if ar.Info[fname].FileSize > 0 then
              arlist.Add(fname);
          arlist.Sort(StrColl);
          if Last then
            LoadLastFileFromArchive(ar, arlist)
          else
            LoadFirstFileFromArchive(ar, arlist);
        end
        else begin
          if isstdin then
          begin
            // stdin
            fsize := len;
            while True do
            begin
              ReallocMem(buf, fsize + StdinBufSize);
              try
                len := Read(buf[fsize], StdinBufSize);
              except
                on E: EIOError do Break;
              end;
              if len = 0 then
                Break;
              Inc(fsize, len);
            end;
          end else
          begin
            // normal file
            fsize := Size;
            ReallocMem(buf, fsize);
            Read(buf[len], fsize - len);
          end; 
          Image.ReadMem(buf, fsize, filter as TImgFilter);
          ArFileList.Clear;
        end;
      finally
        if not (filter is TArchive) then
          filter.Free;
      end;
    finally
      FreeMem(buf);
      Free;
      SetCursor(oldcur);
    end;

  // loding succeeded
  UpdateView(zoom);
end;

procedure TNkVMainForm.Zoom(Value: Double);
//var
//  cw, ch: Integer;
begin
  Stretch := False;
  DispWidth := RoundOff(Image.Width * Value);
  DispHeight := RoundOff(Image.Height * Value);
  //if not FullScreen then
  //  SetClientSize(DispWidth, DispHeight);
  //GetClientSize(cw, ch);
  //if DispWidth > cw then
  //  ViewLeft := (DispWidth - cw) div 2;
  //if DispHeight > ch then
  //  ViewTop := (DispHeight - ch) div 2;
  Scroll(0, 0); // update view
  UpdateCaption;
  Invalidate;
end;

procedure TNkVMainForm.Scroll(X, Y: Integer);
var
  cw, ch, vw, vh: Integer;
begin
  if Stretch then
    Exit;

  GetClientSize(cw, ch);
  if DispWidth < cw then
    vw := DispWidth
  else
    vw := cw;
  if DispHeight < ch then
    vh := DispHeight
  else
    vh := ch;

  Inc(ViewLeft, X);
  if ViewLeft < 0 then
  begin
    Dec(X, ViewLeft);
    ViewLeft := 0;
  end
  else if ViewLeft > (DispWidth-vw) then
  begin
    Inc(X, DispWidth-vw-ViewLeft);
    ViewLeft := DispWidth-vw;
  end;

  Inc(ViewTop, Y);
  if ViewTop < 0 then
  begin
    Dec(Y, ViewTop);
    ViewTop := 0;
  end
  else if ViewTop > (DispHeight-vh) then
  begin
    Inc(Y, DispHeight-vh-ViewTop);
    ViewTop := DispHeight-vh;
  end;

  ScrollWindowEx(Handle, -X, -Y, nil, nil, 0, nil, SW_INVALIDATE);
  Update;
end;

procedure TNkVMainForm.SetClipboardText(const S: string);
var
  h: HGLOBAL;
begin
  h := GlobalAlloc(GHND, Length(S)+1);
  Move(S[1], PChar(GlobalLock(h))^, Length(S));
  GlobalUnlock(h);
  if OpenClipboard(Handle) then
  begin
    EmptyClipboard;
    SetClipboardData(CF_TEXT, h);
    CloseClipBoard;
  end else
    GlobalFree(h);
end;

procedure TNkVMainForm.SaveImage;
var
  fname, ext: string;
begin
  SplitExt(BaseName(FileList[FileList.Cur]), fname, ext);
  if FileList.Count = 0 then
    fname := 'stdin.bmp'
  else
    fname := fname + '.bmp';
  if ShowSaveDlg(Handle, ['All Files', '*.*'], fname) then
    try
      Image.Write(fname, DispWidth, DispHeight, ScaleFilter);
    except
      on E: EIOError do Error(E.Path, E.Msg);
    end;
end;

procedure TNkVMainForm.ShowAboutPlugins;
var
  i: Integer;
  s: string;
begin
  s := IStr(Plugins.Count) + ' plugin(s)'+EOL+EOL;
  for i := 0 to Plugins.Count-1 do
    with TSpiFilter.CreateByFile(Plugins[i]) do
      if Handle = 0 then
        s := s + '(not found)' + EOL
      else
        s := s + FileName + ': ' + Name + EOL;
  MsgBox(Handle, AppName + ' - About Plugins', s);
end;

procedure TNkVMainForm.NextImage(const List: TFileList);
var
  i: Integer;
begin
  i := list.Cur+1;
  while i < list.Count do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list[i])
      else
        LoadFile(list[i], False)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        list.Delete(i);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Continue;
      end;
    end;
    Break;
  end;
  if i < list.Count then
    list.Cur := i;
  UpdateCaption;
  UpdateInfo;
end;

procedure TNkVMainForm.PrevImage(const List: TFileList; Last: Boolean);
var
  i: Integer;
begin
  i := list.Cur-1;
  while i >= 0 do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list[i])
      else
        LoadFile(list[i], Last)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        FileList.Delete(i);
        Dec(i);
        Dec(FileList.Cur);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Dec(i);
        Dec(list.Cur);
        Continue;
      end;
    end;
    Break;
  end;
  if i >= 0 then
    list.Cur := i;
  UpdateCaption;
  UpdateInfo;
end;

procedure TNkVMainForm.UpdateInfo;
begin
  InfoDlg.Path := FileList[FileList.Cur];
  if ArFileList.Count = 0 then
  begin
    InfoDlg.PathInArchive := '';
    if InfoDlg.Path <> '' then
      FileTime(InfoDlg.Path, InfoDlg.ATime, InfoDlg.MTime, InfoDlg.CTime);
  end
  else begin
    InfoDlg.PathInArchive := ArFileList[ArFileList.Cur];
    InfoDlg.MTime := Archiv[InfoDlg.PathInArchive].DateTime;
    FillChar(InfoDlg.CTime, SizeOf(InfoDlg.Ctime), 0);
    FillChar(InfoDlg.ATime, SizeOf(InfoDlg.Atime), 0);
  end;
  InfoDlg.Image := Image;

  if IsWindow(InfoDlg.Handle) then
    InfoDlg.UpdateInfo;
end; 

procedure TNkVMainForm.DoCmd(const Name: string);
const
  keyscroll = 16;
var
  i, w, h: Integer;
begin
  if Name = cmStayOnTop then
    SetStayOnTop(not StayOnTop)
  else if Name = cmHideBorder then
  begin
    if not FullScreen then 
    begin
      Border := not Border;
      SetBorder(Border);
    end;
    UpdateCursor;
  end
  else if Name = cmScaleFilter then
  begin
    ScaleFilter := not ScaleFilter;
    Invalidate;
  end
  else if Name = cmShowGrid then
  begin
    ShowGrid := not ShowGrid;
    Invalidate;
  end
  else if Name = cmNextWindow then
    ActivateNextForm(False)
  else if Name = cmPrevWindow then
    ActivateNextForm(True)
  else if Name = cmAboutPlugins then
    ShowAboutPlugins
  else if Name = cmAbout then
    AboutBox(Handle, AppName, About)
  else if Name = cmExit then
    Close;
 
  if not Image.Loaded then
    Exit;


  if Name = cmFitWindow then
  begin
    if not FullScreen then
    begin
      if Maximized then Restore;
      SetClientSize(Dw, Dh);
    end;
  end
  else if Name = cmFullScreen then SetFullScreen(not FullScreen)
  else if Name = cmZoomIn then
  begin
    for i := 0 to High(ZoomTable) do
      if DispWidth < RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
  end
  else if Name = cmZoomOut then
  begin
    for i := High(ZoomTable) downto 0 do
      if DispWidth > RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
  end
  else if Name = cmZoom50 then Zoom(0.5)
  else if Name = cmZoom100 then Zoom(1.0)
  else if Name = cmZoom150 then Zoom(1.5)
  else if Name = cmZoom200 then Zoom(2.0)
  else if Name = cmZoom300 then Zoom(3.0)
  else if Name = cmZoom400 then Zoom(4.0)
  else if Name = cmZoom800 then Zoom(8.0)
  else if Name = cmStretchMode then
  begin
    Stretch := not Stretch;
    UpdateCaption;
    Invalidate;
  end
  else if Name = cmSetWallpaper then SetWallpaper(False)
  else if Name = cmSetWallpaperTile then SetWallpaper(True)
  else if Name = cmScrollLeft   then Scroll(-keyscroll, 0)
  else if Name = cmScrollRight  then Scroll(keyscroll, 0)
  else if Name = cmScrollUp     then Scroll(0, -keyscroll)
  else if Name = cmScrollDown   then Scroll(0, keyscroll)
  else if Name = cmScrollTop    then Scroll(0, -DispHeight)
  else if Name = cmScrollBottom then Scroll(0, DispHeight)
  else if Name = cmScrollTopBottom then
  begin
    GetClientSize(w, h);
    if Viewtop = DispHeight - h then
      Scroll(0, -DispHeight)
    else
      Scroll(0, DispHeight);
  end
  else if Name = cmShowColorValue then
  begin
    ShowColorValue := not ShowColorValue;
    UpdateCaption;
    UpdateCursor;
  end
  else if Name = cmFirstFile then
  begin
    if DirMode then
      FillFileList;
    FileList.Cur := -1;
    NextImage(FileList);
  end
  else if Name = cmLastFile then
  begin
    if DirMode then
      FillFileList;
    FileList.Cur := FileList.Count;
    PrevImage(FileList, False);
  end
  else if Name = cmNextFile then
  begin
    if DirMode then
      FillFileList;
    NextImage(FileList);
  end
  else if Name = cmPrevFile then
  begin
    if DirMode then
      FillFileList;
    PrevImage(FileList, False);
  end
  else if Name = cmNextImage then
  begin
    if (ArFileList.Count = 0) or (ArFileList.Cur = ArFilelist.Count-1) then
    begin
      if DirMode then
        FillFileList;
      NextImage(FileList);
    end else
      NextImage(ArFileList);
  end
  else if Name = cmPrevImage then
  begin
    if (ArFileList.Count = 0) or (ArFileList.Cur = 0) then
    begin
      if DirMode then
        FillFileList;
      PrevImage(FileList, True);
    end else
      PrevImage(ArFileList, True);
  end
  else if Name = cmCopyFileName then
    SetClipboardText(FileList[FileList.Cur])
  else if Name = cmSaveAs then
    SaveImage
  else if Name = cmInfo then
  begin
    UpdateInfo;
    if not IsWindow(InfoDlg.Handle) then
      InfoDlg.ShowModeless;
  end
  else if Name = cmReloadImage then
    LoadFile(FileList[FileList.Cur], False)
  else if Name = cmRefreshFileList then
  begin
    if DirMode then
      FileListUpdate := True; 
  end;
end;

procedure TNkVMainForm.UpdateCaption;
var
  fname, info: string;
  pt: TPoint;
  r, g, b: Integer;
  zoom: Double;
begin
  if not Image.Loaded then
    Exit;
  zoom := DispWidth / Image.Width;
  if FileList[FileList.Cur] = '' then
    fname := '<STDIN>'
  else
    fname := BaseName(FileList[FileList.Cur]);
  if ArFileList.Count > 0 then
    fname := fname + '/' + ArFileList[ArFileList.Cur];
  if ShowColorValue then
  begin
    GetCursorPos(pt);
    ScreenToClient(pt.x, pt.y, pt.x, pt.y);
    pt.x := Trunc((pt.x+ViewLeft) / zoom);
    pt.y := Trunc((pt.y+ViewTop) / zoom);
    if (pt.x < Integer(Image.Width)) and (pt.y < Integer(Image.Height))
      and (pt.x >= 0) and (pt.y >= 0) then
    begin
      Image.GetPixel(pt.x, pt.y, r, g, b);
      info := ' [R'+IZStr(r, 3)+' G'+IZStr(g, 3)+' B'+IZStr(b, 3)+'] '
              + '(X'+IStr(pt.x)+' Y'+IStr(pt.y)+') ';
    end else
      info := ' [R--- G--- B---] (X--- Y---) ';
  end else
    info := ' (' + IStr(Image.Width) + 'x' + IStr(Image.Height) + 'x' +
      IStr(Image.Bpp) + ') ';

  if Stretch then
    info := info + 'Stretch Mode'
  else
    info := info + IStr(RoundOff(zoom * 100))+'%';
  //info := info + ' ' + IStr(FileList.Cur+1) + '/' + IStr(FileList.Count);
  SetText(fname + ' ' + info  + ' - ' + AppName);
end;

procedure TNkVMainForm.Error(const Path: string; const Msg: string);
var
  s: string;
begin
  if Path = '' then
    s := Msg
  else
    s := Path + EOL + EOL + Msg;
  MessageBox(Handle, PChar(s), PChar(AppName), MB_ICONEXCLAMATION);
end;

procedure TNkVMainForm.SetWallpaper(Tile: Boolean);
var
  key: HKEY;
  data: array[0..MAX_PATH] of Char;
  fname: string;
begin
  fname := JoinPath(ProfDir, 'wallpaper.bmp');
  try
    Image.Write(fname, Image.Width, Image.Height, False);
    if Tile then
      data := '1'
    else
      data := '0';
    if RegOpenKeyEx(HKEY_CURRENT_USER, 'Control Panel\desktop', 0, KEY_WRITE,
      key) = ERROR_SUCCESS then
    begin
      RegSetValueEx(key, 'TileWallPaper', 0, REG_SZ, @data, 1);
      RegCloseKey(key);
    end;
    SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(fname),
      SPIF_UPDATEINIFILE or SPIF_SENDWININICHANGE);
  except
    on E: EIOError do Error(E.Path, E.Msg);
  end;
end;

function EnumWindowsProc(hWnd: HWND; lParam: LPARAM): BOOL; stdcall;
var
  buf: array[0..255] of Char;
begin
  GetClassName(hWnd, buf, SizeOf(buf));
  with TNkvMainForm(lParam) do
    if buf = string(ClassName) then
      FormList.Add(Pointer(hWnd));
  Result := True;
end;

function Compare(Item1, Item2: Pointer): Integer;
begin
  if Integer(Item1) > Integer(Item2) then
    Result := 1
  else if Integer(Item1) < Integer(Item2) then
    Result := -1
  else
    Result := 0;
end;

procedure TNkVMainForm.ActivateNextForm(Prev: Boolean);
var
  i, cur: Integer;
begin
  FormList := TList.Create;
  EnumWindows(@EnumWindowsProc, LPARAM(Self));
  FormList.Sort(Compare);
  cur := 0;
  for i := 0 to FormList.Count-1 do
  begin
    if HWND(FormList[i]) = Handle then
    begin
      cur := i;
      Break;
    end;
  end;
  if Prev then
  begin
    Dec(cur);
    if cur < 0 then
      cur := FormList.Count-1;
  end else
  begin
    Inc(cur);
    if cur = FormList.Count then
      cur := 0;
  end;
  SetForegroundWindow(HWND(FormList[cur]));
  FormList.Free;
end;

procedure TNkVMainForm.WMPaint(var Message: TWMPaint);
var
  dc: HDC;
  ps: TPaintStruct;
  vl, vt, vw, vh, cw, ch, l, t: Integer;
  br: HBRUSH;
begin
  dc := BeginPaint(Handle, ps);

  br := GetStockObject(GRAY_BRUSH);

  GetClientSize(cw, ch);

  if not Image.Loaded then
  begin
    FillRect(dc, Rect(0, 0, cw, ch), br);
    Exit;
  end;

  if Stretch then
  begin
    vl := 0;
    vt := 0;
    if Integer(Image.Width) * ch > Integer(Image.Height) * cw then
    begin
      Dw := cw;
      Dh := RoundOff(Integer(Image.Height) * Dw / Image.Width);
    end else
    begin
      Dh := ch;
      Dw := RoundOff(Integer(Image.Width) * Dh / Image.Height);
    end;
  end else
  begin
    vl := ViewLeft;
    vt := ViewTop;
    Dw := DispWidth;
    Dh := DispHeight;
  end;

  if Dw < cw then
    vw := Dw 
  else
    vw := cw;
  if Dh < ch then
    vh := Dh 
  else
    vh := ch;
  
  if vw < cw then
    l := (cw - vw) div 2
  else
    l := 0;
  if vh < ch then
    t := (ch - vh) div 2
  else
    t := 0;

  Image.Draw(dc, l, t, vw, vh, vl, vt, Dw, Dh,
    ShowGrid and (Dw div Image.Width >= 4), ScaleFilter);

  if vw < cw then
  begin
    FillRect(dc, Rect(l+vw, t, cw, t+vh), br); // right
    FillRect(dc, Rect(0, t, l, t+vh), br); // left
  end;
  if vh < ch then
  begin
    FillRect(dc, Rect(0, t+vh, cw, ch), br); // bottom
    FillRect(dc, Rect(0, 0, cw, t), br); // top
  end;

  EndPaint(Handle, ps);
end;

procedure TNkVMainForm.ShowPluginConfig(const FileName: string);
begin
  with TSpiFilter.CreateByFile(FileName) do
    try
      if Handle = 0 then
      begin
        Error(FileName, 'Load failed.');
        Exit;
      end;
      try
        ShowConfigDlg(Self.Handle);
      except
        on E: EImgFilterError do Error(FileName, E.Msg);
      end;
    finally
      Free;
    end;
end;

procedure TNkVMainForm.WMCommand(var Message: TWMCommand);
begin
  with Message do
    case ItemID of
      IDM_PLUGINSETTING_FIRST .. IDM_PLUGINSETTING_LAST:
        ShowPluginConfig(Plugins[ItemID - IDM_PLUGINSETTING_FIRST]);
    else
      DoCmd(Cmd.GetName(ItemID));
    end;
end;

procedure TNkVMainForm.WMLButtonDown(var Message: TWMLButtonDown);
begin
  if Border or FullScreen or (not Border and (GetKeyState(VK_SHIFT) < 0)) then
  begin
    SetCapture(Handle);
    if Image.Loaded then
      SetCursor(GrabCursor);
    GetCursorPos(OldPt);
  end else
  begin
    ReleaseCapture;
    SendMsg(WM_NCLBUTTONDOWN, HTCAPTION, 0);
  end;
end;

procedure TNkVMainForm.WMLButtonUp(var Message: TWMLButtonUp);
begin
  ReleaseCapture;
  //SetCursor(NormalCursor);
end;

procedure TNkVMainForm.WMMouseMove(var Message: TWMMouseMove);
var
  pt: TPoint;
begin
  if Message.Keys and MK_LBUTTON <> 0 then
  begin
    // If mouse cursor are on outside window,
    // coordinates included in Message are invalid. So use GetCursorPos()
    GetCursorPos(pt);
    Scroll(OldPt.x - pt.x, OldPt.y - pt.y);
    OldPt := pt;
  end
  else if ShowColorValue then
    UpdateCaption;
end;

procedure TNkVMainForm.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  DoCmd(cmFitWindow);
end;

procedure TNkVMainForm.WMContextMenu(var Message: TMessage);
var
  i, id, x, y: Integer;
begin
  ScreenToClient(Message.LParamLo, Message.LParamHi, x, y);
  if y < 0 then
  begin
    inherited;
    Exit;
  end;
  with TPopupMenu.Create do
    try
      CreateCmdMenu(Handle, Cmd, MenuItems);
      if Image.Loaded then
      begin
        if FileList.Count = 0 then // stdin
        begin
          GrayItem(Cmd.GetId(cmPrevImage));
          GrayItem(Cmd.GetId(cmNextImage));
          GrayItem(Cmd.GetId(cmPrevFile));
          GrayItem(Cmd.GetId(cmNextFile));
          GrayItem(Cmd.GetId(cmCopyFileName));
          GrayItem(Cmd.GetId(cmReloadImage));
        end;
        id := Cmd.GetId(cmPluginMenu);
        for i := 0 to Plugins.Count-1 do
          InsertItem(id, BaseName(Plugins[i]), IDM_PLUGINSETTING_FIRST+i);
        DeleteItem(id);
      end else
      begin
        GrayItem(Cmd.GetId(cmPrevImage));
        GrayItem(Cmd.GetId(cmNextImage));
        GrayItem(Cmd.GetId(cmPrevFile));
        GrayItem(Cmd.GetId(cmNextFile));
        GrayItem(Cmd.GetId(cmFirstFile));
        GrayItem(Cmd.GetId(cmLastFile));
        GrayItem(Cmd.GetId(cmInfo));
        GrayItem(Cmd.GetId(cmStretchMode));
        GrayItem(Cmd.GetId(cmFullScreen));
        GrayItem(Cmd.GetId(cmFitWindow));
        GrayItem(Cmd.GetId(cmZoom50));
        GrayItem(Cmd.GetId(cmZoom100));
        GrayItem(Cmd.GetId(cmZoom150));
        GrayItem(Cmd.GetId(cmZoom200));
        GrayItem(Cmd.GetId(cmZoom300));
        GrayItem(Cmd.GetId(cmZoom400));
        GrayItem(Cmd.GetId(cmZoom800));
        GrayItem(Cmd.GetId(cmZoomIn));
        GrayItem(Cmd.GetId(cmZoomOut));
        GrayItem(Cmd.GetId(cmShowColorValue));
        GrayItem(Cmd.GetId(cmCopyFileName));
        GrayItem(Cmd.GetId(cmReloadImage));
        GrayItem(Cmd.GetId(cmSaveAs));
        GrayItem(Cmd.GetId(cmSetWallpaper));
        GrayItem(Cmd.GetId(cmSetWallpaperTile));
        GrayItem(Cmd.GetId(cmPluginMenu));
      end; 
      if FullScreen then
        GrayItem(Cmd.GetId(cmFitWindow));

      CheckItem(Cmd.GetId(cmFullScreen), FullScreen);
      CheckItem(Cmd.GetId(cmStayOnTop), StayOnTop);
      CheckItem(Cmd.GetId(cmStretchMode), Stretch);
      CheckItem(Cmd.GetId(cmHideBorder), not Border);
      CheckItem(Cmd.GetId(cmShowColorValue), ShowColorValue);
      CheckItem(Cmd.GetId(cmScaleFilter), ScaleFilter);
      CheckItem(Cmd.GetId(cmShowGrid), ShowGrid);
      if Message.LParam = $FFFFFFFF then
      begin
        ClientToScreen(0, 0, x, y);
        Popup(Self.Handle, x, y);
      end else
        Popup(Self.Handle, Message.LParamLo, Message.LParamHi);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.WMDropFiles(var Message: TWMDropFiles);
var
  n, i: Integer;
  path, fname: string;
  list: TFileList;
begin
  list := TFileList.Create;
  n := DragQueryFile(Message.Drop, $FFFFFFFF, nil, 0);
  for i := 0 to n-1 do
  begin
    SetLength(path, DragQueryFile(Message.Drop, i, nil, 0));
    // +1 for null terminate
    DragQueryFile(Message.Drop, i, PChar(path), Length(path)+1);
    if IsDir(path) then
    begin
      DirMode := n = 1;
      with TListDir.Create(path) do
        try
          while Each(fname) do
            if not IsDir then
              list.Add(JoinPath(path, fname));
        finally
          Free;
        end
    end else
      list.Add(path);
  end;
  DragFinish(Message.Drop);
  list.Sort(StrColl);
  LoadFirstFile(list);
  list.Free;
  if (FileList.Count = 1) and (FileList[0] <> '') then // not stdin 
  begin
    DirMode := True;
    FileListUpdate := True;
  end;
end;

procedure TNkVMainForm.WMMouseWheel(var Message: TMessage);
begin
  with Message do
    if WParam > 0 then
      DoCmd(cmPrevImage)
    else
      DoCmd(cmNextImage);
end;

procedure TNkVMainForm.WMClose(var Message: TWMClose);
begin
  try
    SaveFormPos(ProfDir, Self);
  except
    on E: EIOError do ErrMsgBox(AppName, E.Path+EOL+EOL+E.Msg);
  end;
  inherited;
end;

procedure TNkVMainForm.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
  try
    SaveFormPos(ProfDir, Self);
  except
    on E: EIOError do ErrMsgBox(AppName, E.Path+EOL+EOL+E.Msg);
  end;
  Message.Result := Integer(True);
end;

end.
